import warnings
from typing import Any, Dict, List, Optional

import numpy as np
import torch.distributed as dist
import wandb
from accelerate import Accelerator
from tqdm import trange

from agilerl.algorithms import GRPO
from agilerl.algorithms.core.base import RLAlgorithm
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.utils.llm_utils import HuggingFaceGym
from agilerl.utils.utils import (
    aggregate_metrics_across_gpus,
    init_wandb,
    save_llm_checkpoint,
    tournament_selection_and_mutation,
)

InitDictType = Optional[Dict[str, Any]]
PopulationType = List[RLAlgorithm]


def finetune_llm(
    pop: List[GRPO],
    env: HuggingFaceGym,
    init_hp: Optional[Dict[str, Any]] = None,
    save_elite: Optional[bool] = None,
    elite_path: Optional[str] = None,
    wb: bool = False,
    evo_steps: Optional[int] = 20,
    tournament: Optional[TournamentSelection] = None,
    mutation: Optional[Mutations] = None,
    wandb_api_key: Optional[str] = None,
    evaluation_interval: int = 10,
    max_reward: Optional[int] = None,
    verbose: bool = True,
    accelerator: Optional[Accelerator] = None,
    max_steps: Optional[int] = None,
):
    """
    Finetunes a population of GRPOs on a HuggingFaceGym environment.

    :param pop: Population of GRPOs to finetune
    :type pop: list[GRPO]
    :param env: HuggingFaceGym environment to finetune on
    :type env: HuggingFaceGym
    :param init_hp: Initial hyperparameters for the population
    :type init_hp: dict, optional
    :param save_elite: Whether to save the elite model, defaults to None
    :type save_elite: bool, optional
    :param elite_path: Path to save the elite model, defaults to None
    :type elite_path: str, optional
    :param wb: Whether to use Weights and Biases, defaults to False
    :type wb: bool, optional
    :param evo_steps: Number of steps between evolution, defaults to None
    :type evo_steps: int, optional
    :param tournament: Tournament selection object, defaults to None
    :type tournament: TournamentSelection, optional
    :param mutation: Mutation object, defaults to None
    :type mutation: Mutations, optional
    :param wandb_api_key: Wandb API key, defaults to None
    :type wandb_api_key: str, optional
    :param evaluation_interval: Number of steps between evaluation, defaults to 10
    :type evaluation_interval: int, optional
    :param max_reward: Maximum reward to aim for, defaults to None
    :type max_reward: int, optional
    :param verbose: Whether to print verbose output, defaults to True
    :type verbose: bool, optional
    :param accelerator: Accelerator object, defaults to None
    :type accelerator: Accelerator, optional
    :param max_steps: Maximum number of steps to run, defaults to None
    :type max_steps: int, optional
    """

    if evo_steps is not None and (tournament is None or mutation is None):
        warnings.warn(
            "'evo_steps' is set but at least one of 'tournament' or 'mutation' is set to None. Evolution will not take place."
        )

    if (tournament is not None and mutation is not None) and evo_steps is None:
        raise ValueError(
            "'evo_steps' must be set if 'tournament' and 'mutation' are not None."
        )

    if mutation is not None:
        assert (
            mutation.architecture_mut == 0
        ), "Architecture mutation is not allowed for LLM finetuning."
        assert (
            mutation.new_layer_prob == 0
        ), "New layer mutation is not allowed for LLM finetuning."
        assert (
            mutation.parameters_mut == 0
        ), "Network parameters mutation is not allowed for LLM finetuning."
        assert (
            mutation.activation_mut == 0
        ), "Activation mutation is not allowed for LLM finetuning."

    if init_hp is None:
        init_hp = {}
        init_hp["BATCH_SIZE_PER_GPU"] = pop[0].batch_size
        init_hp["ALGO"] = pop[0].algo
    data_increment = (
        getattr(dist, "get_world_size", lambda: 1)() if dist.is_initialized() else 1
    )
    grad_accum = getattr(pop[0].actor, "gradient_accumulation_steps", lambda: 1)()
    effective_data_batch_size = data_increment * env.data_batch_size_per_gpu
    effective_learning_batch_size = (
        data_increment * init_hp["BATCH_SIZE_PER_GPU"] * grad_accum
    )
    if accelerator is None or accelerator.is_main_process:
        print(
            f"""
=========================================================================
Commencing RL finetuning

Data batch size per gpu: {env.data_batch_size_per_gpu}
Number of GPUs: {data_increment}
Gradient accumulation: {grad_accum}
Effective data batch size: {data_increment} * {env.data_batch_size_per_gpu} = {effective_data_batch_size}
Effective learning batch_size: {data_increment} * {init_hp["BATCH_SIZE_PER_GPU"]} * {grad_accum} = {effective_learning_batch_size}
=========================================================================
        """
        )
    if wb and (accelerator is None or accelerator.is_main_process):
        init_hp["effective_data_batch_size"] = effective_data_batch_size
        init_hp["effective_learning_batch_size"] = effective_learning_batch_size
        init_hp["distributed_training"] = True if accelerator is not None else False
        init_wandb(
            algo=init_hp["ALGO"],
            env_name=env.name,
            wandb_api_key=wandb_api_key,
            init_hyperparams=init_hp,
        )
    if accelerator is None or accelerator.is_main_process:
        print("\nTraining...")

    bar_format = "{l_bar}{bar:10}| {n:4}/{total_fmt} [{elapsed:>7}<{remaining:>7}, {rate_fmt}{postfix}]"
    if max_steps is None:
        max_steps = len(env)
    training_steps = max_steps // effective_data_batch_size
    if accelerator is None or accelerator.is_main_process:
        pbar = trange(
            max_steps,
            unit="step",
            bar_format=bar_format,
            ascii=True,
            dynamic_ncols=True,
        )

    total_steps = 0
    # calling env.reset() supplies the first batch of training data
    prompts = env.reset(reset_dataloaders=True)
    for i in range(training_steps):
        agent_metrics_dict = {}
        for agent_idx, agent in enumerate(pop):
            completion_ids, action_masks = agent.get_action(prompts)
            completion_lengths = np.mean([x.shape[1] for x in completion_ids])

            # Use the reward function stored in env.step to calculate reward of the each answer from the group
            next_prompts, rewards = env.step(completion_ids)
            experiences = (
                completion_ids,
                action_masks,
                rewards,
            )
            loss, kl = agent.learn(experiences)
            metrics = [loss, kl, rewards, completion_lengths]
            if max_reward is not None:
                accuracy = (rewards == max_reward).sum() / len(rewards.flatten())
                metrics.append(accuracy)
            agg_metrics = [
                aggregate_metrics_across_gpus(accelerator, metric) for metric in metrics
            ]
            prompts = next_prompts
            agent.steps[-1] += effective_data_batch_size
            total_steps += effective_data_batch_size
            agg_test_metrics = None
            if (i + 1) % evaluation_interval == 0:
                test_reward = agent.test(env)
                test_metrics = [test_reward]
                if max_reward is not None:
                    test_accuracy = (test_reward == max_reward).sum() / len(
                        test_reward.flatten()
                    )
                    test_metrics.append(test_accuracy)
                agg_test_metrics = [
                    aggregate_metrics_across_gpus(accelerator, metric)
                    for metric in test_metrics
                ]

                if accelerator is not None:
                    accelerator.wait_for_everyone()
            if accelerator is None or accelerator.is_main_process:
                metrics_dict = {
                    "Train/Loss": agg_metrics[0],
                    "Train/KL-divergence": agg_metrics[1],
                    "Train/Mean reward": agg_metrics[2],
                    "Train/Average completion length": int(agg_metrics[3]),
                }
                if max_reward is not None:
                    metrics_dict |= {"Train/Accuracy": agg_metrics[4]}
                agent_metrics_dict[f"agent_{agent_idx}/train_metrics"] = metrics_dict
                if agg_test_metrics is not None:
                    test_metrics_dict = {"Eval/Mean reward": agg_test_metrics[0]}
                    if max_reward is not None:
                        test_metrics_dict |= {"Eval/Accuracy": agg_test_metrics[1]}
                    agent_metrics_dict[f"agent_{agent_idx}/test_metrics"] = (
                        test_metrics_dict
                    )
                pbar.update(effective_data_batch_size)
                agent.scores.append(agg_metrics[2])

        if accelerator is not None:
            accelerator.wait_for_everyone()
        if tournament and mutation is not None:
            if (i + 1) % evo_steps == 0:
                if accelerator is not None:
                    accelerator.wait_for_everyone()
                pop = tournament_selection_and_mutation(
                    population=pop,
                    tournament=tournament,
                    mutation=mutation,
                    env_name=env.name,
                    accelerator=accelerator,  # Set as None for LLM finetuning as it does not require the same accelerator handling as standard RL models
                    language_model=True,
                    elite_path=elite_path,
                    save_elite=save_elite,
                )
                if accelerator is not None:
                    accelerator.wait_for_everyone()
        else:
            if (i + 1) % max_steps == 0:
                save_llm_checkpoint(agent, elite_path, i + 1)

        if wb and (accelerator is None or accelerator.is_main_process):
            wandb_dict = {
                "Train/Best reward": np.max(
                    [
                        agent_metrics_dict[f"agent_{agent_idx}/train_metrics"][
                            "Train/Mean reward"
                        ]
                        for agent_idx, _ in enumerate(pop)
                    ]
                ),
                "Train/Mean population reward": np.mean(
                    [
                        agent_metrics_dict[f"agent_{agent_idx}/train_metrics"][
                            "Train/Mean reward"
                        ]
                        for agent_idx, _ in enumerate(pop)
                    ]
                ),
                "Train/Mean population loss": np.mean(
                    [
                        agent_metrics_dict[f"agent_{agent_idx}/train_metrics"][
                            "Train/Loss"
                        ]
                        for agent_idx, _ in enumerate(pop)
                    ]
                ),
                "Train/Mean population KL divergence": np.mean(
                    [
                        agent_metrics_dict[f"agent_{agent_idx}/train_metrics"][
                            "Train/KL-divergence"
                        ]
                        for agent_idx, _ in enumerate(pop)
                    ]
                ),
                "Train/Mean population completion length": np.mean(
                    [
                        agent_metrics_dict[f"agent_{agent_idx}/train_metrics"][
                            "Train/Average completion length"
                        ]
                        for agent_idx, _ in enumerate(pop)
                    ]
                ),
            }
            if max_reward is not None:
                wandb_dict |= {
                    "Train/Mean population accuracy": np.mean(
                        [
                            agent_metrics_dict[f"agent_{agent_idx}/train_metrics"][
                                "Train/Accuracy"
                            ]
                            for agent_idx, _ in enumerate(pop)
                        ]
                    )
                }
            if len(pop[0].registry.hp_config.config.keys()) > 0:
                wandb_dict |= {
                    f"HPO_agent_{agent_idx}/{key}": getattr(agent, key)
                    for agent_idx, agent in enumerate(pop)
                    for key in agent.registry.hp_config.config.keys()
                }

            if agg_test_metrics is not None:
                test_dict = {
                    "Eval/Best reward": np.max(
                        [
                            agent_metrics_dict[f"agent_{agent_idx}/test_metrics"][
                                "Eval/Mean reward"
                            ]
                            for agent_idx, _ in enumerate(pop)
                        ]
                    ),
                    "Eval/Mean population reward": np.mean(
                        [
                            agent_metrics_dict[f"agent_{agent_idx}/test_metrics"][
                                "Eval/Mean reward"
                            ]
                            for agent_idx, _ in enumerate(pop)
                        ]
                    ),
                }
                if max_reward is not None:
                    test_dict |= {
                        "Eval/Mean population accuracy": np.mean(
                            [
                                agent_metrics_dict[f"agent_{agent_idx}/test_metrics"][
                                    "Eval/Accuracy"
                                ]
                                for agent_idx, _ in enumerate(pop)
                            ]
                        )
                    }
                wandb_dict |= test_dict
            wandb.log(wandb_dict)

    if (
        verbose
        and total_steps > evaluation_interval
        and (accelerator is None or accelerator.is_main_process)
    ):
        fitness = [str(round(agent.fitness[-1], 2)) for agent in pop]
        avg_fitness = ["%.2f" % np.mean(agent.fitness[-5:]) for agent in pop]
        avg_score = ["%.2f" % np.mean(agent.scores[-10:]) for agent in pop]
        agents = [agent.index for agent in pop]
        num_steps = [agent.steps[-1] for agent in pop]
        muts = [agent.mut for agent in pop]
        print(
            f"""
            --- Global Steps {total_steps} ---
            Fitness:\t\t{fitness}
            Score:\t\t{agg_metrics[2]}
            5 fitness avgs:\t{avg_fitness}
            10 score avgs:\t{avg_score}
            Agents:\t\t{agents}
            Steps:\t\t{num_steps}
            Mutations:\t\t{muts}
            """,
            end="\r",
        )

    if accelerator is not None:
        accelerator.wait_for_everyone()
    if accelerator is None or accelerator.is_main_process:
        pbar.close()
        if wb:
            wandb.finish()



import warnings
from collections import OrderedDict
from numbers import Number
from typing import Iterator

import numpy as np
import torch
from tensordict import TensorDict, tensorclass
from torch.utils.data import IterableDataset

from agilerl.components import ReplayBuffer
from agilerl.typing import ArrayOrTensor, ObservationType


def to_tensordict(
    data: ObservationType, dtype: torch.dtype = torch.float32
) -> TensorDict:
    """Converts a tuple or dict of torch.Tensor or np.ndarray to a TensorDict.

    :param data: Tuple or dict of torch.Tensor or np.ndarray.
    :type data: ObservationType
    :param dtype: Data type of the TensorDict, defaults to torch.float32
    :type dtype: torch.dtype, optional
    :return: TensorDict, whether the data was a tuple or not.
    """
    if isinstance(data, tuple):
        assert all(
            isinstance(el, (torch.Tensor, np.ndarray, Number)) for el in data
        ), "Expected all elements of the tuple to be torch.Tensor or np.ndarray."

        new_data = OrderedDict()
        for i, el in enumerate(data):
            new_data[f"tuple_obs_{i}"] = el

        data = TensorDict(new_data)

    elif isinstance(data, dict):
        assert all(
            isinstance(el, (torch.Tensor, np.ndarray, Number)) for el in data.values()
        ), "Expected all values of the dict to be torch.Tensor or np.ndarray."

        data = TensorDict(data)

    return data.to(dtype=dtype)


def to_torch_tensor(data: ArrayOrTensor, dtype=torch.float32) -> torch.Tensor:
    """Converts a numpy array or Python number to a torch tensor.

    :param data: Numpy array or Python number.
    :type data: ArrayOrTensor
    :param dtype: Data type of the torch tensor, defaults to torch.float32
    :type dtype: torch.dtype, optional
    :return: Torch tensor.
    """
    if isinstance(data, (np.ndarray, Number, bool, np.bool_)):
        return torch.tensor(data, dtype=dtype)
    elif isinstance(data, torch.Tensor):
        return data.to(dtype=dtype)
    else:
        # Handle any other types by attempting to convert to tensor
        return torch.tensor(data, dtype=dtype)


@tensorclass
class Transition:
    obs: ObservationType
    action: ArrayOrTensor
    next_obs: ObservationType
    reward: ArrayOrTensor
    done: ArrayOrTensor

    def __post_init__(self) -> None:
        # Convert observations to TensorDict if they are dicts or tuples
        if isinstance(self.obs, (dict, tuple)):
            self.obs = to_tensordict(self.obs)

        if isinstance(self.next_obs, (dict, tuple)):
            self.next_obs = to_tensordict(self.next_obs)

        # Convert all data to torch tensors with proper dtype
        self.action = to_torch_tensor(self.action)
        self.done = to_torch_tensor(self.done)
        self.reward = to_torch_tensor(self.reward)

        if self.done.ndim == 0:
            self.done = self.done.unsqueeze(-1)

        if self.reward.ndim == 0:
            self.reward = self.reward.unsqueeze(-1)


class ReplayDataset(IterableDataset):
    """
    Iterable Dataset containing the ReplayBuffer which will be updated with new
    experiences during training

    :param buffer: Experience replay buffer
    :type buffer: agilerl.components.replay_buffer.ReplayBuffer()
    :param batch_size: Number of experiences to sample at a time, defaults to 256
    :type batch_size: int, optional
    """

    def __init__(self, buffer: ReplayBuffer, batch_size: int = 256) -> None:
        if not isinstance(buffer, ReplayBuffer):
            warnings.warn("Buffer is not an agilerl ReplayBuffer.")

        assert batch_size > 0, "Batch size must be greater than zero."
        self.buffer = buffer
        self.batch_size = batch_size

    def __iter__(self) -> Iterator:
        samples = self.buffer.sample(self.batch_size)
        yield samples



"""An async vector pettingzoo environment"""

import multiprocessing as mp
import sys
import time
import traceback
from collections import OrderedDict, defaultdict
from copy import deepcopy
from enum import Enum
from multiprocessing.connection import Connection
from multiprocessing.sharedctypes import RawArray
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeAlias, TypeVar, Union

import numpy as np
from gymnasium import logger, spaces
from gymnasium.error import (
    AlreadyPendingCallError,
    ClosedEnvironmentError,
    NoAsyncCallError,
)
from gymnasium.vector.utils import CloudpickleWrapper, clear_mpi_env_vars
from pettingzoo import ParallelEnv

from agilerl.typing import ActionType, GymSpaceType, MultiAgentStepType, NumpyObsType
from agilerl.vector.pz_vec_env import PettingZooVecEnv

AgentID = TypeVar("AgentID")
ObsType = TypeVar("ObsType")
PzEnvType = Union[PettingZooVecEnv, ParallelEnv]
SharedMemoryType: TypeAlias = Union[RawArray, Tuple[RawArray, ...], Dict[str, RawArray]]


def reshape_observation(
    raw_data: NumpyObsType, space: spaces.Space, num_envs: int
) -> Any:
    """Reshape the raw data to the correct shape for the observation space.

    :param raw_data: The raw data to reshape
    :type raw_data: np.ndarray, Dict[str, np.ndarray], Tuple[np.ndarray, ...]
    :param space: The observation space
    :type space: gymnasium.spaces.Space
    :param num_envs: The number of environments
    :type num_envs: int

    :return: The reshaped data
    :rtype: Any
    """
    if isinstance(space, spaces.Dict):
        result = OrderedDict()
        for key, subspace in space.spaces.items():
            result[key] = reshape_observation(raw_data[key], subspace, num_envs)

    elif isinstance(space, spaces.Tuple):
        result = []
        for i, subspace in enumerate(space.spaces):
            result.append(reshape_observation(raw_data[i], subspace, num_envs))

        result = tuple(result)
    else:
        # Reshape to [num_envs, *shape]
        shape = space.shape if space.shape != () else (1,)
        reshaped = raw_data.reshape((num_envs, *shape))
        result = reshaped.astype(space.dtype)

    return result


class AsyncState(Enum):
    """The AsyncVectorEnv possible states given the different actions."""

    DEFAULT = "default"
    WAITING_RESET = "reset"
    WAITING_STEP = "step"
    WAITING_CALL = "call"


class AsyncPettingZooVecEnv(PettingZooVecEnv):
    """Vectorized PettingZoo environment that runs multiple environments in parallel

    :param env_fns: Functions that create the environment
    :type env_fns: list[Callable]
    :param copy: Boolean flag to copy the observation data when it is returned with either .step() or .reset(), recommended, defaults to True
    :type copy: bool, optional
    :param context: Context for multiprocessing

    """

    processes: List[mp.Process]
    parent_pipes: List[Connection]
    error_queue: mp.Queue
    _state: AsyncState

    def __init__(
        self,
        env_fns: List[Callable[[], PzEnvType]],
        copy: bool = True,
        context: Optional[str] = None,
    ):
        # Core class attributes
        ctx = mp.get_context(context)
        self.env_fns = env_fns
        self.num_envs = len(env_fns)
        dummy_env = env_fns[0]()
        self.metadata = (
            dummy_env.metadata
            if hasattr(dummy_env, "metadata")
            else dummy_env.unwrapped.metadata
        )
        self.render_mode = (
            dummy_env.render_mode
            if hasattr(dummy_env, "render_mode")
            else dummy_env.unwrapped.render_mode
        )
        self.possible_agents = dummy_env.possible_agents
        self.active_agents = None
        self.previous_active = None
        self.copy = copy
        self.agents = dummy_env.possible_agents[:]
        action_spaces = {
            agent: dummy_env.action_space(agent) for agent in dummy_env.possible_agents
        }
        observation_spaces = {
            agent: dummy_env.observation_space(agent)
            for agent in dummy_env.possible_agents
        }
        dummy_env.close()
        del dummy_env
        super().__init__(
            len(env_fns),
            observation_spaces,
            action_spaces,
            self.possible_agents,
        )

        # Create the shared memory for sharing observations between subprocesses
        self._obs_buffer = create_shared_memory(
            num_envs=self.num_envs,
            obs_spaces=self._single_observation_spaces,
            context=ctx,
        )
        self.observations = Observations(
            shared_memory=self._obs_buffer,
            obs_spaces=self._single_observation_spaces,
            num_envs=self.num_envs,
        )
        self.parent_pipes, self.processes = [], []
        self.error_queue = ctx.Queue()

        target = _async_worker
        with clear_mpi_env_vars():
            for idx, env_fn in enumerate(self.env_fns):
                parent_pipe, child_pipe = ctx.Pipe()
                process = ctx.Process(
                    target=target,
                    name=f"Worker<{type(self).__name__}>-{idx}",
                    args=(
                        idx,
                        CloudpickleWrapper(env_fn),
                        child_pipe,
                        parent_pipe,
                        self._obs_buffer,
                        self.error_queue,
                        self.agents,
                    ),
                )
                self.parent_pipes.append(parent_pipe)
                self.processes.append(process)
                process.daemon = True
                process.start()
                child_pipe.close()

        self._state = AsyncState.DEFAULT

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> Tuple[Dict[str, NumpyObsType], Dict[str, Any]]:
        """
        Reset all the environments and return two dictionaries of batched observations and infos.

        :param seed: Random seed, defaults to None
        :type seed: None | int, optional
        :param options: Options dictionary
        :type options: dict[str, Any]
        """
        self.reset_async(seed=seed, options=options)
        return self.reset_wait()

    def reset_async(
        self,
        seed: Optional[int] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Send calls to the :obj:`reset` methods of the sub-environments.

        To get the results of these calls, you may invoke :meth:`reset_wait`.

        :param seed: Random seed, defaults to None
        :type seed: None | int, optional
        :param options: Options dictionary
        :type options: dict[str, Any]
        """
        self._assert_is_running()
        if seed is None:
            seed = [None for _ in range(self.num_envs)]
        elif isinstance(seed, int):
            seed = [seed + i for i in range(self.num_envs)]
        assert (
            len(seed) == self.num_envs
        ), f"If seeds are passed as a list the length must match num_envs={self.num_envs} but got length={len(seed)}."

        if self._state != AsyncState.DEFAULT:
            raise AlreadyPendingCallError(
                f"Calling `reset_async` while waiting for a pending call to `{self._state.value}` to complete",
                str(self._state.value),
            )
        for pipe, env_seed in zip(self.parent_pipes, seed):
            env_kwargs = {"seed": env_seed, "options": options}
            pipe.send(("reset", env_kwargs))

        self._state = AsyncState.WAITING_RESET

    def get_observations(self) -> Dict[str, NumpyObsType]:
        """Get the observations from the environments.

        :return: Observations from the environments
        :rtype: Dict[str, NumpyObsType]
        """
        return (
            {
                agent: deepcopy(self.observations[agent])
                for agent in self.observations.keys()
            }
            if self.copy
            else self.observations
        )

    def reset_wait(
        self, timeout: Optional[float] = None
    ) -> Tuple[Dict[str, NumpyObsType], Dict[str, Any]]:
        """Waits for the calls triggered by :meth:`reset_async` to finish and returns the results.

        :param timeout: Number of seconds before the call to ``reset_wait`` times out. If `None`, the call to
        ``reset_wait`` never times out, defaults to 0
        :type timeout: int | float | None, optional

        :return: Tuple of observations and infos
        :rtype: Tuple[Dict[str, NumpyObsType], Dict[str, Any]]
        """
        self._assert_is_running()
        if self._state != AsyncState.WAITING_RESET:
            raise NoAsyncCallError(
                "Calling `reset_wait` without any prior " "call to `reset_async`.",
                AsyncState.WAITING_RESET.value,
            )

        if not self._poll_pipe_envs(timeout):
            self._state = AsyncState.DEFAULT
            raise mp.TimeoutError(
                f"The call to `reset_wait` has timed out after {timeout} second(s)."
            )

        info_data, successes = zip(*[pipe.recv() for pipe in self.parent_pipes])
        self._raise_if_errors(successes)

        infos = {}
        # Convert info data to list before passing into key_search function
        for i, info in enumerate(info_data):
            infos = self._add_info(infos, info, i)

        self._state = AsyncState.DEFAULT

        return (
            self.get_observations(),
            infos,
        )

    def step_async(self, actions: List[List[ActionType]]) -> None:
        """
        Tell all the environments to start taking a step with the given actions.
        Call step_wait() to get the results of the step.
        You should not call this if a step_async run is already pending.

        :param actions: List of lists of length num_envs, each sub list contains actions for each agent in a given environment
        :type actions: list[list[int | float | np.ndarray]]
        """
        self._assert_is_running()
        if self._state != AsyncState.DEFAULT:
            raise AlreadyPendingCallError(
                f"Calling `step_async` while waiting for a pending call to `{self._state.value}` to complete.",
                str(self._state.value),
            )
        for pipe, action in zip(self.parent_pipes, actions):
            pipe.send(("step", action))

        self._state = AsyncState.WAITING_STEP

    def step_wait(self, timeout: Optional[float] = None) -> MultiAgentStepType:
        """
        Wait for the calls to :obj:`step` in each sub-environment to finish.

        :param timeout: Number of seconds before the call to ``step_wait`` times out. If `None`, the call to ``step_wait`` never times out, defaults to 0
        :type timeout: int | float | None, optional

        :return: Tuple of observations, rewards, dones, and infos
        :rtype: Tuple[Dict[str, NumpyObsType], Dict[str, np.ndarray], Dict[str, np.ndarray], Dict[str, Any]]
        """

        self._assert_is_running()
        if self._state != AsyncState.WAITING_STEP:
            raise NoAsyncCallError(
                "Calling `step_wait` without any prior call " "to `step_async`.",
                AsyncState.WAITING_STEP.value,
            )

        if not self._poll_pipe_envs(timeout):
            self._state = AsyncState.DEFAULT
            raise mp.TimeoutError(
                f"The call to `step_wait` has timed out after {timeout} second(s)."
            )

        rewards, terminations, truncations, infos = (
            defaultdict(list) for _ in range(4)
        )
        successes = []
        infos = {}
        for env_idx, pipe in enumerate(self.parent_pipes):
            env_step_return, success = pipe.recv()
            successes.append(success)
            if success:
                reward, term, trunc, info = env_step_return
                for agent in self.agents:
                    rewards[agent].append(reward[agent])
                    terminations[agent].append(term[agent])
                    truncations[agent].append(trunc[agent])

                infos = self._add_info(infos, info, env_idx)

        self._raise_if_errors(successes)
        self._state = AsyncState.DEFAULT

        rewards = {agent: np.array(rew) for agent, rew in rewards.items()}
        terminations = {agent: np.array(term) for agent, term in terminations.items()}
        truncations = {agent: np.array(trunc) for agent, trunc in truncations.items()}

        return (
            self.get_observations(),
            rewards,
            terminations,
            truncations,
            infos,
        )

    def render(self) -> Any:
        """
        Returns the rendered frames from the parallel environments.

        """
        return self.call("render")

    def call(self, name: str, *args: Any, **kwargs: Any) -> Any:
        """
        Call a method from each parallel environment with args and kwargs.

        :param name: Name of the method or property to call
        :type name: str
        :param *args: Position arguments to apply to the method call.
        :type *args: Any
        :param **kwargs: Keyword arguments to apply to the method call.
        :type **kwargs: Any
        """
        self.call_async(name, *args, **kwargs)
        return self.call_wait()

    def call_async(self, name: str, *args: Any, **kwargs: Any) -> None:
        """
        Calls the method with name asynchronously and apply args and kwargs to the method.

        :param name: Name of the method or property to call
        :type name: str
        :param *args: Position arguments to apply to the method call.
        :type *args: Any
        :param **kwargs: Keyword arguments to apply to the method call.
        :type **kwargs: Any
        """
        self._assert_is_running()
        if self._state != AsyncState.DEFAULT:
            raise AlreadyPendingCallError(
                f"Calling `call_async` while waiting for a pending call to `{self._state.value}` to complete.",
                str(self._state.value),
            )

        for pipe in self.parent_pipes:
            pipe.send(("_call", (name, args, kwargs)))

        self._state = AsyncState.WAITING_CALL

    def call_wait(self, timeout: Optional[float] = None) -> Any:
        """Calls all parent pipes and waits for the results.

        :param timeout: Number of seconds before the call to :meth:`call_wait` times out. If ``None`` (default),
        the call to :meth:`call_wait` never times out, defaults to 0
        :type timeout: int | float | None, optional
        """
        self._assert_is_running()
        if self._state != AsyncState.WAITING_CALL:
            raise NoAsyncCallError(
                "Calling `call_wait` without any prior call to `call_async`.",
                AsyncState.WAITING_CALL.value,
            )

        if not self._poll_pipe_envs(timeout):
            self._state = AsyncState.DEFAULT
            raise mp.TimeoutError(
                f"The call to `call_wait` has timed out after {timeout} second(s)."
            )

        results, successes = zip(*[pipe.recv() for pipe in self.parent_pipes])
        self._raise_if_errors(successes)
        self._state = AsyncState.DEFAULT
        return results

    def get_attr(self, name: str) -> Any:
        """
        Get a property from each parallel environment.

        :param name: Name of property to get from each individual environment
        :type name: str
        """
        return self.call(name)

    def set_attr(self, name: str, values: Any) -> None:
        """Sets an attribute of the sub-environments.

        :param name: Name of the property to be set in each individual environment.
        :type name: str
        :param values: Values of the property to be set to. If ``values`` is a list or
            tuple, then it corresponds to the values for each individual
            environment, otherwise a single value is set for all environments.
        :type values: list[Any] | tuple[Any] | object
        """
        self._assert_is_running()
        if not isinstance(values, (list, tuple)):
            values = [values for _ in range(self.num_envs)]
        if len(values) != self.num_envs:
            raise ValueError(
                "Values must be a list or tuple with length equal to the number of environments. "
                f"Got `{len(values)}` values for {self.num_envs} environments."
            )

        if self._state != AsyncState.DEFAULT:
            raise AlreadyPendingCallError(
                f"Calling `set_attr` while waiting for a pending call to `{self._state.value}` to complete.",
                str(self._state.value),
            )

        for pipe, value in zip(self.parent_pipes, values):
            pipe.send(("_setattr", (name, value)))
        _, successes = zip(*[pipe.recv() for pipe in self.parent_pipes])
        self._raise_if_errors(successes)

    def close_extras(
        self, timeout: Optional[float] = None, terminate: bool = False
    ) -> None:
        """
        Close the environments & clean up the extra resources (processes and pipes).

        :param timeout: Number of seconds before the call to :meth:`close` times out. If ``None``,
                the call to :meth:`close` never times out. If the call to :meth:`close`
                times out, then all processes are terminated, defaults to 0
        :type timeout: int | float | None, optional
        :param terminate: If ``True``, then the :meth:`close` operation is forced and all processes are terminated, defaults to False
        :type terminate: bool, optional
        """
        timeout = 0 if terminate else timeout

        try:
            if self._state != AsyncState.DEFAULT:
                logger.warn(
                    f"Calling `close` while waiting for a pending call to `{self._state.value}` to complete."
                )
                function = getattr(self, f"{self._state.value}_wait")
                function(timeout)
        except mp.TimeoutError:
            terminate = True

        if terminate:
            for process in self.processes:
                if process.is_alive():
                    process.terminate()
        else:
            for pipe in self.parent_pipes:
                if (pipe is not None) and (not pipe.closed):
                    pipe.send(("close", None))

            for pipe in self.parent_pipes:
                if (pipe is not None) and (not pipe.closed):
                    pipe.recv()

        for pipe in self.parent_pipes:
            if pipe is not None:
                pipe.close()
        for process in self.processes:
            process.join()

    def _poll_pipe_envs(self, timeout: Optional[float] = None) -> bool:
        self._assert_is_running()

        if timeout is None:
            return True

        end_time = time.perf_counter() + timeout
        for pipe in self.parent_pipes:
            delta = max(end_time - time.perf_counter(), 0)

            if pipe is None:
                return False
            if pipe.closed or (not pipe.poll(delta)):
                return False
        return True

    def _raise_if_errors(self, successes: List[bool]) -> None:
        if all(successes):
            return

        num_errors = self.num_envs - sum(successes)
        assert num_errors > 0
        for i in range(num_errors):
            index, exctype, value, trace = self.error_queue.get()

            logger.error(
                f"Received the following error from Worker-{index} - Shutting it down"
            )
            logger.error(f"{trace}")

            self.parent_pipes[index].close()
            self.parent_pipes[index] = None

            if i == num_errors - 1:
                logger.error("Raising the last exception back to the main process.")
                self._state = AsyncState.DEFAULT
                raise exctype(value)

    def _assert_is_running(self) -> None:
        if self.closed:
            raise ClosedEnvironmentError(
                f"Trying to operate on `{type(self).__name__}`, after a call to `close()`."
            )

    def _add_info(
        self, vector_infos: Dict[str, Any], env_info: Dict[str, Any], env_num: int
    ) -> Dict[str, Any]:
        """
        Compile a vectorized information dictionary
        """

        for key, value in env_info.items():
            # If value is a dictionary, then we apply the `_add_info` recursively.
            if isinstance(value, dict):
                array = self._add_info(vector_infos.get(key, {}), value, env_num)
            # Otherwise, we are a base case to group the data
            else:
                # If the key doesn't exist in the vector infos, then we can create an array of that batch type
                if key not in vector_infos:
                    if type(value) in [int, float, bool] or issubclass(
                        type(value), np.number
                    ):
                        array = np.zeros(self.num_envs, dtype=type(value))
                    elif isinstance(value, np.ndarray):
                        # We assume that all instances of the np.array info are of the same shape
                        array = np.zeros(
                            (self.num_envs, *value.shape), dtype=value.dtype
                        )
                    elif value is None:
                        array = np.full(
                            self.num_envs, fill_value=np.nan, dtype=np.float32
                        )
                    else:
                        # For unknown objects, we use a Numpy object array
                        array = np.full(self.num_envs, fill_value=None, dtype=object)

                # Otherwise, just use the array that already exists
                else:
                    array = vector_infos[key]

                # Assign the data in the `env_num` position
                #   We only want to run this for the base-case data (not recursive data forcing the ugly function structure)
                array[env_num] = value

            # Get the array mask and if it doesn't already exist then create a zero bool array
            array_mask = vector_infos.get(
                f"_{key}", np.zeros(self.num_envs, dtype=np.bool_)
            )
            array_mask[env_num] = True

            # Update the vector info with the updated data and mask information
            vector_infos[key], vector_infos[f"_{key}"] = array, array_mask

        return vector_infos

    def __del__(self) -> None:
        """On deleting the object, checks that the vector environment is closed."""
        if not getattr(self, "closed", True) and hasattr(self, "_state"):
            self.close(terminate=True)
        if hasattr(self, "_obs_buffer"):
            del self._obs_buffer
        if hasattr(self, "observations"):
            del self.observations


class Observations:
    """
    Class for storing observations with a dictionary interface

    :param shared_memory: A RawArray that all envs write observations to.
    :type shared_memory: multiprocessing.RawArray
    :param obs_spaces: Dictionary of gymnasium observation spaces
    :type obs_spaces: Dict[str, gymnasiums.spaces.Space]
    :param num_envs: Number of environments
    :type num_envs: int
    """

    obs_view: Dict[str, NumpyObsType]

    def __init__(
        self,
        shared_memory: Dict[str, SharedMemoryType],
        obs_spaces: Dict[str, spaces.Space],
        num_envs: int,
    ):
        self.num_envs = num_envs
        self.shared_memory = shared_memory
        self.obs_view = {}
        self.agents = list(self.shared_memory.keys())
        self.obs_spaces = obs_spaces
        for agent, shm in shared_memory.items():
            agent_space = obs_spaces[agent]
            if isinstance(agent_space, spaces.Dict):
                obs_view = OrderedDict()
                for key, subspace in agent_space.spaces.items():
                    obs_view[key] = np.frombuffer(
                        shm[key].get_obj(), dtype=subspace.dtype
                    )
                self.obs_view[agent] = obs_view
            elif isinstance(agent_space, spaces.Tuple):
                obs_view = tuple(
                    np.frombuffer(shm[i].get_obj(), dtype=agent_space.spaces[i].dtype)
                    for i in range(len(agent_space.spaces))
                )
            else:
                obs_view = np.frombuffer(shm.get_obj(), dtype=agent_space.dtype)

            self.obs_view[agent] = obs_view

    def __getitem__(self, agent: str) -> NumpyObsType:
        """
        Get agent observation given a key (agent_id)
        """
        return reshape_observation(
            self.obs_view[agent], self.obs_spaces[agent], self.num_envs
        )

    def __str__(self) -> str:
        return f"{self.obs_view}"

    def __repr__(self) -> str:
        return self.__str__()

    def __contains__(self, key: str) -> bool:
        return key in self.obs_view.keys()

    def __len__(self) -> int:
        return len(self.agents)

    def __iterate_kv(self) -> Any:
        for key in self.obs_view:
            yield (key, self.__getitem__(key))

    def __iter__(self) -> Any:
        return self.__iterate_kv()

    def keys(self) -> Any:
        for k, _ in self.__iterate_kv():
            yield k

    def values(self) -> Any:
        for _, v in self.__iterate_kv():
            yield v

    def items(self) -> Any:
        return self.__iterate_kv()

    def get(self, key: str) -> Optional[np.ndarray]:
        try:
            return self.__getitem__(key)
        except KeyError:
            return None


def _create_memory_array(
    num_envs: int, obs_space: spaces.Space, context: Any
) -> RawArray:
    """Create a shared memory array for a given observation space.

    :param num_envs: Number of environments
    :type num_envs: int
    :param obs_space: Observation space
    :type obs_space: gymnasium.spaces.Space
    :param context: Multiprocessing context
    :type context: Any
    """
    return context.Array(obs_space.dtype.char, num_envs * int(np.prod(obs_space.shape)))


def create_shared_memory(
    num_envs: int, obs_spaces: Dict[str, spaces.Space], context: Any
) -> Dict[str, RawArray]:
    """
    Create shared memory for multi-agent observations.

    :param num_envs: Number of environments
    :type num_envs: int
    :param obs_spaces: Dictionary of gymnasium observation spaces
    :type obs_spaces: Dict[str, gymnasiums.spaces.Space]
    :param context: Multiprocessing context
    :type context: Any
    """
    shared_memory = {}
    for agent, obs_space in obs_spaces.items():
        if isinstance(obs_space, spaces.Dict):
            shm = OrderedDict()
            for key, subspace in obs_space.spaces.items():
                shm[key] = _create_memory_array(num_envs, subspace, context)
        elif isinstance(obs_space, spaces.Tuple):
            shm = tuple(
                _create_memory_array(num_envs, subspace, context)
                for subspace in obs_space.spaces
            )
        else:
            shm = _create_memory_array(num_envs, obs_space, context)

        shared_memory[agent] = shm

    return shared_memory


def get_placeholder_value(
    agent: str,
    transition_name: str,
    obs_spaces: Optional[Dict[str, spaces.Space]] = None,
) -> Any:
    """Used to obtain a placeholder value to return for associated experience when an
    agent is killed or is inactive for the current step.

    :param agent: Agent ID
    :type agent: str
    :param transition_name: Name of the transition
    :type transition_name: str
    :param obs_spaces: Observation spaces
    :type obs_spaces: Dict[str, gymnasium.spaces.Space]

    :return: Placeholder value
    :rtype: Any
    """
    match transition_name:
        case "reward":
            return np.nan
        case "truncated":
            return np.nan
        case "terminated":
            return np.nan
        case "info":
            return {}
        case "observation":
            if obs_spaces is None:
                return None

            agent_space = obs_spaces[agent]
            if isinstance(agent_space, spaces.Dict):
                # For Dict spaces, create a dictionary of -1 arrays
                return {k: np.full(v.shape, np.nan) for k, v in agent_space.items()}
            elif isinstance(agent_space, spaces.Tuple):
                # For Tuple spaces, create a tuple of -1 arrays
                return tuple(np.full(s.shape, np.nan) for s in agent_space)
            else:
                # For normal spaces
                return np.full(agent_space.shape, np.nan)


def process_transition(
    transitions: Tuple[Dict[str, NumpyObsType], ...],
    obs_spaces: Dict[str, spaces.Space],
    transition_names: List[str],
    agents: List[str],
) -> List[Dict[str, NumpyObsType]]:
    """Process transition, adds in placeholder values for killed sub-agents.

    :param transitions: Tuple of environment transition
    :type transitions: Tuple[Dict[str, NumpyObsType], ...]
    :param obs_spaces: Observation spaces
    :type obs_spaces: Dict[str, gymnasium.spaces.Space]
    :param transition_names: Names associated to transitions
    :type transition_names: List[str]
    :param agents: List of sub-agent names
    :type agents: List[str]
    """
    transition_list = []
    for transition, name in zip(list(transitions), transition_names):
        transition = {
            agent: (
                transition[agent]
                if agent in transition.keys()
                else get_placeholder_value(agent, name, obs_spaces)
            )
            for agent in agents
        }
        transition_list.append(transition)
    return transition_list


def write_vector_observation(
    index: int,
    observation: np.ndarray,
    shared_memory: RawArray,
    obs_space: spaces.Space,
) -> None:
    """Write a vector observation to the shared memory.

    :param index: Environment index
    :type index: int
    :param observation: Observation from env.step or env.reset
    :type observation: np.ndarray
    :param shared_memory: Shared memory
    :type shared_memory: multiprocessing.RawArray
    :param obs_space: Observation space
    :type obs_space: gymnasium.spaces.Space
    """
    size = int(np.prod(obs_space.shape))
    dtype = obs_space.dtype
    dest = np.frombuffer(shared_memory.get_obj(), dtype=dtype)
    np.copyto(
        dest[index * size : (index + 1) * size],
        np.asarray(observation, dtype=dtype).flatten(),
    )


def write_to_shared_memory(
    index: int,
    observation: Dict[str, NumpyObsType],
    shared_memory: Dict[str, RawArray],
    obs_space: Dict[str, GymSpaceType],
) -> None:
    """Set the observation for a given environment. Handles Dict and Tuple spaces.

    :param index: Environment index
    :type index: int
    :param observation: Observation from env.step or env.reset
    :type observation: Dict[str, np.ndarray]
    :param shared_memory: Shared memory
    :type shared_memory: Dict[str, mp.Array | Tuple[mp.Array, ...] | Dict[str, mp.Array]]
    :param obs_space: Observation space dictionary
    :type obs_space: Dict[str, gymnasium.spaces.Space]
    """
    for agent in shared_memory.keys():
        agent_space = obs_space[agent]
        obs = observation[agent]
        if isinstance(agent_space, spaces.Dict):
            for key, subspace in agent_space.spaces.items():
                write_vector_observation(
                    index, obs[key], shared_memory[agent][key], subspace
                )

        elif isinstance(agent_space, spaces.Tuple):
            for i, subspace in enumerate(agent_space.spaces):
                write_vector_observation(
                    index, obs[i], shared_memory[agent][i], subspace
                )
        else:
            write_vector_observation(index, obs, shared_memory[agent], agent_space)


def _async_worker(
    index: int,
    env_fn: Callable[[], ParallelEnv],
    pipe: Connection,
    parent_pipe: Connection,
    shared_memory: Dict[str, RawArray],
    error_queue: mp.Queue,
    agents: List[str],
) -> None:
    """
    Worker function to run the environment in a subprocess.

    :param index: Subprocess index
    :type index: int
    :param env_fn: Function to call environment
    :type env_fn: callable
    :param pipe: Child pipe object for sending data to the main process
    :type pipe: Connection
    :param parent_pipe: Parent pipe object
    :type parent_pipe: Connection
    :param shared_memory: List of shared memories.
    :type shared_memory: List[multiprocessing.Array]
    :param error_queue: Queue object for collecting subprocess errors to communicate back to the main process
    :type error_queue: mp.Queue
    :param observation_shapes: Shapes of observations
    :type observation_shapes: Dict[str, Tuple[int]]
    :param observation_widths: Flattened observation widths
    :type observation_widths: Dict[str, int]
    :param observation_dtypes: Observation dtypes
    :type observation_dtypes: Dict[str, np.dtype]
    :param agents: Sub-agent names
    :type agents: str

    """
    env = env_fn()
    observation_space = {agent: env.observation_space(agent) for agent in agents}
    parent_pipe.close()

    # Need to keep track of the active agents in the environment
    current_active = None
    try:
        while True:
            command, data = pipe.recv()
            if command == "reset":
                obs, info = env.reset(**data)
                current_active = list(obs.keys())
                transition = obs, info
                observation, info = process_transition(
                    transition,
                    observation_space,
                    ["observation", "info"],
                    agents,
                )
                write_to_shared_memory(
                    index, observation, shared_memory, observation_space
                )
                pipe.send((info, True))
            elif command == "step":
                data = {
                    possible_agent: (
                        np.array(data[idx]).squeeze()
                        if not isinstance(data[idx], int)
                        else data[idx]
                    )
                    for idx, possible_agent in enumerate(current_active)
                }
                observation, reward, terminated, truncated, info = env.step(data)
                if all(
                    [
                        term | trunc
                        for term, trunc in zip(terminated.values(), truncated.values())
                    ]
                ):
                    observation, info = env.reset()

                current_active = list(observation.keys())
                transition = observation, reward, terminated, truncated, info
                observation, reward, terminated, truncated, info = process_transition(
                    transition,
                    observation_space,
                    [
                        "observation",
                        "reward",
                        "terminated",
                        "truncated",
                        "info",
                    ],
                    agents,
                )
                write_to_shared_memory(
                    index, observation, shared_memory, observation_space
                )
                pipe.send(((reward, terminated, truncated, info), True))

            elif command == "close":
                pipe.send((None, True))
                break
            elif command == "_call":
                name, args, kwargs = data
                if name in ["reset", "step", "close", "_setattr", "_check_spaces"]:
                    raise ValueError(
                        f"Trying to call function `{name}` with `call`, use `{name}` directly instead."
                    )
                attr = getattr(env, name)
                if callable(attr):
                    pipe.send((attr(*args, **kwargs), True))
                else:
                    pipe.send((attr, True))
            elif command == "_setattr":
                name, value = data
                setattr(env, name, value)
                pipe.send((None, True))
            else:
                raise RuntimeError(
                    f"Received unknown command `{command}`. Must be one of [`reset`, `step`, `close`, `_call`, `_setattr`, `_check_spaces`]."
                )

    except (KeyboardInterrupt, Exception):
        error_type, error_message, _ = sys.exc_info()
        trace = traceback.format_exc()
        error_queue.put((index, error_type, error_message, trace))
        pipe.send((None, False))

    finally:
        env.close()



import math
import warnings
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn

from agilerl.modules.base import EvolvableModule, MutationType, mutation


class EvolvableBERT(EvolvableModule):
    """The Evolvable BERT class.

    :param encoder_layers: Encoder layer(s) hidden size
    :type encoder_layers: list[int]
    :param decoder_layers: Decoder layer(s) hidden size
    :type decoder_layers: list[int]
    :param end2end: End to end transformer, using positional and token embeddings, defaults to True
    :type end2end: bool, optional
    :param src_vocab_size: Source vocabulary size, defaults to 10837
    :type src_vocab_size: int, optional
    :param tgt_vocab_size: Target vocabulary size, defaults to 10837
    :type tgt_vocab_size: int, optional
    :param encoder_norm: Encoder output normalization, defaults to True
    :type encoder_norm: bool, optional
    :param decoder_norm: Decoder output normalization, defaults to True
    :type decoder_norm: bool, optional
    :param d_model: Number of expected features in the encoder/decoder inputs, defaults to 512
    :type d_model: int, optional
    :param n_head: Number of heads in the multiheadattention models, defaults to 8
    :type n_head: int, optional
    :param dropout: Dropout value, defaults to 0.1
    :type dropout: float, optional
    :param activation: Activation function of encoder/decoder intermediate layer, defaults to 'ReLU'
    :type activation: str, optional
    :param layer_norm_eps: Epsilon value in layer normalization components, defaults to 1e-5
    :type layer_norm_eps: float, optional
    :param batch_first: Input/output tensor order. True:(batch, seq, feat.) False:(seq, batch, feat.). Defaults to False
    :type batch_first: bool, optional
    :param norm_first: Perform LayerNorm before other attention and feedforward operations, defaults to False
    :type norm_first: bool, optional
    :param max_encoder_layers: Maximum number of encoder layers, defaults to 12
    :type max_encoder_layers: int, optional
    :param max_decoder_layers: Maximum number of decoder layers, defaults to 12
    :type max_decoder_layers: int, optional
    :param device: Device for accelerated computing, 'cpu' or 'cuda', defaults to 'cpu'
    :type device: str, optional
    """

    def __init__(
        self,
        encoder_layers: List[int],
        decoder_layers: List[int],
        end2end: bool = True,
        src_vocab_size: int = 10837,
        tgt_vocab_size: int = 10837,
        encoder_norm: bool = True,
        decoder_norm: bool = True,
        d_model: int = 512,
        n_head: int = 8,
        dropout: float = 0.1,
        activation: str = "relu",
        layer_norm_eps: float = 1e-5,
        batch_first: bool = False,
        norm_first: bool = False,
        max_encoder_layers: int = 12,
        max_decoder_layers: int = 12,
        device: str = "cpu",
        name: str = "bert",
    ) -> None:
        super().__init__(device)

        self.encoder_layers = encoder_layers
        self.decoder_layers = decoder_layers
        self.end2end = end2end
        self.src_vocab_size = src_vocab_size
        self.tgt_vocab_size = tgt_vocab_size
        self.encoder_norm = encoder_norm
        self.decoder_norm = decoder_norm
        self.d_model = d_model
        self.n_head = n_head
        self.dropout = dropout
        self.name = name
        self._activation = activation
        self.layer_norm_eps = layer_norm_eps
        self.batch_first = batch_first
        self.norm_first = norm_first
        self.max_encoder_layers = max_encoder_layers
        self.max_decoder_layers = max_decoder_layers

        if self.end2end:
            self.generator = nn.Linear(self.d_model, tgt_vocab_size)
            self.src_tok_emb = TokenEmbedding(src_vocab_size, self.d_model)
            self.tgt_tok_emb = TokenEmbedding(tgt_vocab_size, self.d_model)
            self.positional_encoding = PositionalEncoder(self.d_model, self.dropout)
        else:
            self.wte = TokenEmbedding(src_vocab_size, self.d_model)
            if len(self.encoder_layers) > 0:
                self.wpe = PositionalEncoding(self.d_model, self.encoder_layers[0])
            else:
                self.wpe = PositionalEncoding(self.d_model, self.decoder_layers[0])

        self.encoder, self.decoder = self.build_networks()
        self.encoder_keys = list(self.encoder.keys())
        self.decoder_keys = list(self.decoder.keys())

    @property
    def activation(self) -> str:
        return self._activation

    @activation.setter
    def activation(self, activation: str) -> None:
        self._activation = activation

    def build_networks(self):
        """Creates and returns transformer neural network."""
        encoder_dict = OrderedDict()
        decoder_dict = OrderedDict()

        # Create the encoder
        for n, dim_feedfwd in enumerate(self.encoder_layers):
            encoder_dict[f"{self.name}_encoder_layer_{str(n)}"] = (
                nn.modules.TransformerEncoderLayer(
                    self.d_model,
                    self.n_head,
                    dim_feedfwd,
                    self.dropout,
                    self.activation,
                    self.layer_norm_eps,
                    self.batch_first,
                    self.norm_first,
                    device=self.device,
                )
            )
        if self.encoder_norm:
            encoder_dict[f"{self.name}_encoder_norm_0"] = (
                nn.modules.normalization.LayerNorm(
                    self.d_model, eps=self.layer_norm_eps, device=self.device
                )
            )

        # Create the decoder
        for n, dim_feedfwd in enumerate(self.decoder_layers):
            decoder_dict[f"{self.name}_decoder_layer_{str(n)}"] = (
                nn.modules.TransformerDecoderLayer(
                    self.d_model,
                    self.n_head,
                    dim_feedfwd,
                    self.dropout,
                    self.activation,
                    self.layer_norm_eps,
                    self.batch_first,
                    self.norm_first,
                    device=self.device,
                )
            )
        if self.decoder_norm:
            decoder_dict[f"{self.name}_decoder_norm_0"] = (
                nn.modules.normalization.LayerNorm(
                    self.d_model, eps=self.layer_norm_eps, device=self.device
                )
            )

        self._reset_parameters()

        return nn.ModuleDict(encoder_dict), nn.ModuleDict(decoder_dict)

    def generate_square_subsequent_mask(self, sz):
        """Returns a square mask for the sequence that prevents the model from looking into the future words when
        making predictions.
        The masked positions are filled with float('-inf'). Unmasked positions are filled with float(0.0).

        :param sz: Size of mask to generate
        :type sz: int
        """
        mask = (torch.triu(torch.ones((sz, sz), device=self.device)) == 1).transpose(
            0, 1
        )
        mask = (
            mask.float()
            .masked_fill(mask == 0, float("-inf"))
            .masked_fill(mask == 1, float(0.0))
        )
        return mask

    def create_mask(
        self, src: torch.Tensor, tgt: torch.Tensor, pad_idx: int
    ) -> Tuple[torch.Tensor, ...]:
        """Returns masks to hide source and target padding tokens.

        :param src: Source
        :type src: torch.Tensor
        :param tgt: Target
        :type tgt: torch.Tensor
        :param pad_idx: Index of padding symbol <pad> in special symbols list
        :type pad_idx: int
        """
        src_seq_len = src.shape[0]
        tgt_seq_len = tgt.shape[0]

        tgt_mask = self.generate_square_subsequent_mask(tgt_seq_len)
        src_mask = torch.zeros((src_seq_len, src_seq_len), device=self.device).type(
            torch.bool
        )

        src_padding_mask = (src == pad_idx).transpose(0, 1)
        tgt_padding_mask = (tgt == pad_idx).transpose(0, 1)
        return (
            src_mask.to(self.device),
            tgt_mask.to(self.device),
            src_padding_mask.to(self.device),
            tgt_padding_mask.to(self.device),
        )

    def _reset_parameters(self):
        """Initiate parameters in the transformer model."""
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def forward(
        self,
        src: torch.Tensor,
        tgt: torch.Tensor,
        src_mask: Optional[torch.Tensor] = None,
        tgt_mask: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
        src_key_padding_mask: Optional[torch.Tensor] = None,
        tgt_key_padding_mask: Optional[torch.Tensor] = None,
        memory_key_padding_mask: Optional[torch.Tensor] = None,
        is_causal: bool = False,
    ) -> torch.Tensor:
        """Returns output of neural network.

        :param src: Encoder input sequence
        :type src: torch.Tensor
        :param tgt: Decoder input sequence
        :type tgt: torch.Tensor
        :param src_mask: Additive mask for the src sequence, defaults to None
        :type src_mask: Optional[torch.Tensor], optional
        :param tgt_mask: Additive mask for the tgt sequence, defaults to None
        :type tgt_mask: Optional[torch.Tensor], optional
        :param memory_mask: Additive mask for the encoder output, defaults to None
        :type memory_mask: Optional[torch.Tensor], optional
        :param src_key_padding_mask: Tensor mask for src keys per batch, defaults to None
        :type src_key_padding_mask: Optional[torch.Tensor], optional
        :param tgt_key_padding_mask: Tensor mask for tgt keys per batch, defaults to None
        :type tgt_key_padding_mask: Optional[torch.Tensor], optional
        :param memory_key_padding_mask: Tensor mask for memory keys per batch, defaults to None
        :type memory_key_padding_mask: Optional[torch.Tensor], optional
        :param is_causal: Applies a causal mask as mask and ignores attn_mask for computing scaled dot product attention, defaults to False
        :type is_causal: bool, optional
        """
        encoder_output, encoder_hidden_states = self.encode(
            src, src_mask, src_key_padding_mask, is_causal
        )
        memory = encoder_output
        decoder_output, decoder_hidden_states = self.decode(
            tgt,
            memory,
            tgt_mask,
            memory_mask,
            tgt_key_padding_mask,
            memory_key_padding_mask,
        )

        if self.end2end:
            decoder_output = self.generator(decoder_output)

        return decoder_output

    def encode(
        self,
        src: torch.Tensor,
        src_mask: Optional[torch.Tensor] = None,
        src_key_padding_mask: Optional[torch.Tensor] = None,
        is_causal: bool = False,
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, ...]]:
        """Returns encoded transformer input.

        :param src: Encoder input sequence
        :type src: torch.Tensor
        :param src_mask: Additive mask for the src sequence, defaults to None
        :type src_mask: torch.Tensor, optional
        :param src_key_padding_mask: Tensor mask for src keys per batch, defaults to None
        :type src_key_padding_mask: torch.Tensor, optional
        :param is_causal: Applies a causal mask as mask and ignores attn_mask for computing scaled dot product attention, defaults to False
        :type is_causal: bool, optional
        """
        if self.end2end:
            src = self.positional_encoding(self.src_tok_emb(src))

        # Encoder forward pass preparation
        src_key_padding_mask = _canonical_mask(
            mask=src_key_padding_mask,
            mask_name="src_key_padding_mask",
            other_type=_none_or_dtype(src_mask),
            other_name="mask",
            target_type=src.dtype,
        )
        encoder_output = src
        first_layer = self.encoder[self.encoder_keys[0]]
        str_first_layer = "self.net[0]"
        src_key_padding_mask_for_layers = src_key_padding_mask
        (
            encoder_output,
            convert_to_nested,
            src_key_padding_mask_for_layers,
        ) = self.check_encoder_sparsity_fast_path(
            src,
            encoder_output,
            first_layer,
            str_first_layer,
            src_mask,
            src_key_padding_mask,
            src_key_padding_mask_for_layers,
        )

        # Prevent type refinement
        make_causal = is_causal is True

        if is_causal is None:
            if src_mask is not None:
                sz = src_mask.size(0)
                causal_comparison = torch.triu(
                    torch.ones(sz, sz, device=src_mask.device) * float("-inf"),
                    diagonal=1,
                ).to(src_mask.dtype)
                if torch.equal(src_mask, causal_comparison):
                    make_causal = True
        is_causal = make_causal

        all_hidden_states = ()

        # Encoder forward pass
        for key in self.encoder_keys:
            if "norm" not in key:
                all_hidden_states = all_hidden_states + (encoder_output,)
                encoder_output = self.encoder[key](
                    encoder_output,
                    src_mask=src_mask,
                    is_causal=is_causal,
                    src_key_padding_mask=src_key_padding_mask_for_layers,
                )
        all_hidden_states = all_hidden_states + (encoder_output,)
        if convert_to_nested:
            encoder_output = encoder_output.to_padded_tensor(0.0)
            all_hidden_states = all_hidden_states + (encoder_output,)
        if "encoder_norm_0" in self.encoder_keys:
            encoder_output = self.encoder["encoder_norm_0"](encoder_output)
            all_hidden_states = all_hidden_states + (encoder_output,)
        return encoder_output, all_hidden_states

    def decode(
        self,
        tgt: torch.Tensor,
        memory: torch.Tensor,
        tgt_mask: Optional[torch.Tensor] = None,
        memory_mask: Optional[torch.Tensor] = None,
        tgt_key_padding_mask: Optional[torch.Tensor] = None,
        memory_key_padding_mask: Optional[torch.Tensor] = None,
    ) -> Tuple[torch.Tensor, Tuple[torch.Tensor, ...]]:
        """Returns decoded transformer input.

        :param tgt: Decoder input sequence
        :type tgt: torch.Tensor
        :param memory: Encoder output sequence
        :type memory: torch.Tensory
        :param tgt_mask: Additive mask for the tgt sequence, defaults to None
        :type tgt_mask: torch.Tensor, optional
        :param memory_mask: Additive mask for the encoder output, defaults to None
        :type memory_mask: torch.Tensor, optional
        :param tgt_key_padding_mask: Tensor mask for tgt keys per batch, defaults to None
        :type tgt_key_padding_mask: torch.Tensor, optional
        :param memory_key_padding_mask: Tensor mask for memory keys per batch, defaults to None
        :type memory_key_padding_mask: torch.Tensor, optional
        """
        if self.end2end:
            tgt = self.positional_encoding(self.src_tok_emb(tgt))

        all_hidden_states = ()

        # Decoder forward pass
        decoder_output = tgt
        for key in self.decoder_keys:
            if "norm" not in key:
                all_hidden_states = all_hidden_states + (decoder_output,)
                decoder_output = self.decoder[key](
                    decoder_output,
                    memory,
                    tgt_mask=tgt_mask,
                    memory_mask=memory_mask,
                    tgt_key_padding_mask=tgt_key_padding_mask,
                    memory_key_padding_mask=memory_key_padding_mask,
                )
        all_hidden_states = all_hidden_states + (decoder_output,)
        if "decoder_norm_0" in self.decoder_keys:
            decoder_output = self.decoder["decoder_norm_0"](decoder_output)
            all_hidden_states = all_hidden_states + (decoder_output,)
        return decoder_output, all_hidden_states

    def check_encoder_sparsity_fast_path(
        self,
        src: torch.Tensor,
        output: torch.Tensor,
        first_layer: nn.Module,
        str_first_layer: str,
        mask: torch.Tensor,
        src_key_padding_mask: torch.Tensor,
        src_key_padding_mask_for_layers: torch.Tensor,
    ) -> Tuple[torch.Tensor, bool, torch.Tensor]:
        """Returns encoder output, conversion to nested and padding mask depending on if sparsity fast path possible.
        :param src: Encoder input sequence
        :type src: torch.Tensor
        :param output: Encoder output sequence
        :type output: torch.Tensor
        :param first_layer: First layer of encoder
        :type first_layer: torch.Module()
        :param str_first_layer: Name of first layer of encoder
        :type str_first_layer: str
        :param mask: Mask for the src sequence
        :type mask: torch.Tensor
        :param src_key_padding_mask: Tensor mask for src keys per batch
        :type src_key_padding_mask: torch.Tensor
        :param src_key_padding_mask_for_layers: Tensor mask for src keys per batch for layers
        :type src_key_padding_mask_for_layers: torch.Tensor
        """
        convert_to_nested = False
        if (
            isinstance(first_layer, torch.nn.TransformerEncoderLayer)
            and first_layer.norm_first
            and first_layer.training
            and first_layer.self_attn.batch_first
            and first_layer.self_attn._qkv_same_embed_dim
            and first_layer.norm1.eps == first_layer.norm2.eps
            and src.dim() == 3
            and src_key_padding_mask is not None
            and torch._nested_tensor_from_mask_left_aligned(
                src, src_key_padding_mask.logical_not()
            )
            and not output.is_nested
            and mask is None
            and first_layer.self_attn.num_heads % 2 != 1
            and not torch.is_autocast_enabled()
        ):
            tensor_args = (
                src,
                first_layer.self_attn.in_proj_weight,
                first_layer.self_attn.in_proj_bias,
                first_layer.self_attn.out_proj.weight,
                first_layer.self_attn.out_proj.bias,
                first_layer.norm1.weight,
                first_layer.norm1.bias,
                first_layer.norm2.weight,
                first_layer.norm2.bias,
                first_layer.linear1.weight,
                first_layer.linear1.bias,
                first_layer.linear2.weight,
                first_layer.linear2.bias,
            )

            if (
                (src.is_cuda or "cpu" in str(src.device))
                and torch.is_grad_enabled()
                and any(x.requires_grad for x in tensor_args)
            ):
                convert_to_nested = True
                output = torch._nested_tensor_from_mask(
                    output, src_key_padding_mask.logical_not(), mask_check=False
                )
                src_key_padding_mask_for_layers = None

        return output, convert_to_nested, src_key_padding_mask_for_layers

    def count_parameters(self, without_layer_norm: bool = False) -> int:
        """Returns number of parameters in neural network.

        :param without_layer_norm: Exclude normalization layers, defaults to False
        :type without_layer_norm: bool, optional
        """
        count = 0
        for name, param in self.named_parameters():
            if not without_layer_norm or "layer_norm" not in name:
                count += param.data.cpu().numpy().flatten().shape[0]
        return count

    @mutation(MutationType.LAYER)
    def add_encoder_layer(self):
        """Adds an encoder layer to transformer."""
        if len(self.encoder_layers) < self.max_encoder_layers:
            self.encoder_layers += [self.encoder_layers[-1]]
        # else:
        #     self.add_node()

    @mutation(MutationType.LAYER)
    def add_decoder_layer(self):
        """Adds a decoder layer to transformer."""
        if len(self.decoder_layers) < self.max_decoder_layers:
            self.decoder_layers += [self.decoder_layers[-1]]
        # else:
        #     self.add_node()

    @mutation(MutationType.LAYER)
    def remove_encoder_layer(self):
        """Removes an encoder layer from transformer."""
        if len(self.encoder_layers) > 1:
            self.encoder_layers = self.encoder_layers[:-1]
        # else:
        #     self.add_node()

    @mutation(MutationType.LAYER)
    def remove_decoder_layer(self):
        """Removes a decoder layer from transformer."""
        if len(self.decoder_layers) > 1:
            self.decoder_layers = self.decoder_layers[:-1]
        # else:
        #     self.add_node()

    @mutation(MutationType.NODE)
    def add_node(
        self,
        network: Optional[str] = None,
        hidden_layer: Optional[int] = None,
        numb_new_nodes: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Adds nodes to hidden layer of encoder/decoder.

        :param network: Network to add node to, 'encoder' or 'decoder', defaults to None
        :type network: str, optional
        :param hidden_layer: Depth of hidden layer to add nodes to, defaults to None
        :type hidden_layer: int, optional
        :param numb_new_nodes: Number of nodes to add to hidden layer, defaults to None
        :type numb_new_nodes: int, optional

        :return: Dictionary containing hidden layer, number of new nodes and network
        :rtype: Dict[str, Any]
        """
        if network is None:
            network = np.random.choice(["encoder", "decoder"], 1)[0]

        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([16, 32, 64], 1)[0]

        if network == "encoder":
            if hidden_layer is None:
                hidden_layer = np.random.randint(0, len(self.encoder_layers), 1)[0]
            else:
                hidden_layer = min(hidden_layer, len(self.encoder_layers) - 1)

            self.encoder_layers[hidden_layer] += numb_new_nodes
        else:
            if hidden_layer is None:
                hidden_layer = np.random.randint(0, len(self.decoder_layers), 1)[0]
            else:
                hidden_layer = min(hidden_layer, len(self.decoder_layers) - 1)

            self.decoder_layers[hidden_layer] += numb_new_nodes

        return {
            "hidden_layer": hidden_layer,
            "numb_new_nodes": numb_new_nodes,
            "network": network,
        }

    @mutation(MutationType.NODE)
    def remove_node(
        self,
        network: Optional[str] = None,
        hidden_layer: Optional[int] = None,
        numb_new_nodes: Optional[int] = None,
    ) -> Dict[str, Any]:
        """Removes nodes from hidden layer of encoder/decoder.

        :param network: Network to remove node from, 'encoder' or 'decoder', defaults to None
        :type network: Optional[str], optional
        :param hidden_layer: Depth of hidden layer to remove nodes from, defaults to None
        :type hidden_layer: Optional[int], optional
        :param numb_new_nodes: Number of nodes to remove from hidden layer, defaults to None
        :type numb_new_nodes: Optional[int], optional

        :return: Dictionary containing hidden layer, number of removed nodes and network
        :rtype: Dict[str, Any]
        """
        if network is None:
            network = np.random.choice(["encoder", "decoder"], 1)[0]

        if numb_new_nodes is None:
            numb_new_nodes = np.random.choice([16, 32, 64], 1)[0]

        if network == "encoder":
            if hidden_layer is None:
                hidden_layer = np.random.randint(0, len(self.encoder_layers), 1)[0]

            else:
                hidden_layer = min(hidden_layer, len(self.encoder_layers) - 1)
            if self.encoder_layers[hidden_layer] - numb_new_nodes > 64:  # HARD LIMIT
                self.encoder_layers[hidden_layer] -= numb_new_nodes
        else:
            if hidden_layer is None:
                hidden_layer = np.random.randint(0, len(self.decoder_layers), 1)[0]
            else:
                hidden_layer = min(hidden_layer, len(self.decoder_layers) - 1)

            if self.decoder_layers[hidden_layer] - numb_new_nodes > 64:  # HARD LIMIT
                self.decoder_layers[hidden_layer] -= numb_new_nodes

        return {
            "hidden_layer": hidden_layer,
            "numb_new_nodes": numb_new_nodes,
            "network": network,
        }

    def recreate_network(self) -> None:
        """Recreates neural network."""
        new_encoder, new_decoder = self.build_networks()

        new_encoder = EvolvableModule.preserve_parameters(
            old_net=self.encoder, new_net=new_encoder
        )
        self.decoder = EvolvableModule.preserve_parameters(
            old_net=self.decoder, new_net=new_decoder
        )

        self.encoder = new_encoder
        self.decoder = new_decoder


def _canonical_mask(
    mask: Optional[torch.Tensor],
    mask_name: str,
    other_type: Optional[torch.dtype],
    other_name: str,
    target_type: torch.dtype,
    check_other: bool = True,
) -> Optional[torch.Tensor]:
    """Returns canonical mask. Adapted from torch.nn.functional.

    :param mask: Input mask tensor
    :type mask: Optional[torch.Tensor]
    :param mask_name: Name of the mask
    :type mask_name: str
    :param other_type: Data type of the other tensor
    :type other_type: Optional[torch.dtype]
    :param other_name: Name of the other tensor
    :type other_name: str
    :param target_type: Target data type for the mask
    :type target_type: torch.dtype
    :param check_other: Flag to check other tensor type, defaults to True
    :type check_other: bool, optional
    :return: Canonical mask tensor
    :rtype: Optional[torch.Tensor]
    """
    if mask is not None:
        _mask_dtype = mask.dtype
        _mask_is_float = torch.is_floating_point(mask)
        if _mask_dtype != torch.bool and not _mask_is_float:
            raise AssertionError(
                f"only bool and floating types of {mask_name} are supported"
            )
        if check_other and other_type is not None:
            if _mask_dtype != other_type:
                warnings.warn(
                    f"Support for mismatched {mask_name} and {other_name} "
                    "is deprecated. Use same type for both instead."
                )
        if not _mask_is_float:
            mask = torch.zeros_like(mask, dtype=target_type).masked_fill_(
                mask, float("-inf")
            )
    return mask


class PositionalEncoder(nn.Module):
    """The Positional Encoder class.
    Adds positional encoding to the token embedding to introduce a notion of word order.

    :param emb_size: Number of expected features
    :type emb_size: int
    :param dropout: Dropout value, defaults to 0.1
    :type dropout: float, optional
    :param maxlen: Maximum length of sequence, defaults to 5000
    :type maxlen: int, optional
    """

    pos_embedding: torch.Tensor

    def __init__(self, emb_size: int, dropout: float, maxlen: int = 5000) -> None:
        super().__init__()

        den = torch.exp(-torch.arange(0, emb_size, 2) * math.log(10000) / emb_size)
        pos = torch.arange(0, maxlen).reshape(maxlen, 1)
        pos_embedding = torch.zeros((maxlen, emb_size))
        pos_embedding[:, 0::2] = torch.sin(pos * den)
        pos_embedding[:, 1::2] = torch.cos(pos * den)
        pos_embedding = pos_embedding.unsqueeze(-2)

        self.dropout = nn.Dropout(dropout)
        self.register_buffer("pos_embedding", pos_embedding)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through positional encoder.
        :param x: Input to positional encoder, shape [seq_len, batch_size, embedding_dim]
        :type x: torch.Tensor
        """
        return self.dropout(x + self.pos_embedding[: x.size(0), :])


class PositionalEncoding(nn.Module):
    """The positional embedding class.
    Converts tensor of input indices into corresponding tensor of position embeddings.
    """

    def __init__(self, max_positions: int, emb_size: int):
        super().__init__()
        self.embedding = nn.Embedding(max_positions, emb_size)
        self.emb_size = emb_size

    def forward(self, tokens: torch.Tensor):
        """Forward pass through position embedding module.
        :param tokens: Tokens to embed
        :type tokens: torch.Tensor
        """
        return self.embedding(tokens)


class TokenEmbedding(nn.Module):
    """The token embedding class. Converts tensor of input indices into corresponding
    tensor of token embeddings.

    :param vocab_size: Size of the vocabulary
    :type vocab_size: int
    :param emb_size: Size of the embedding
    :type emb_size: int
    """

    def __init__(self, vocab_size: int, emb_size: int):
        super().__init__()
        self.embedding = nn.Embedding(vocab_size, emb_size)
        self.emb_size = emb_size

    def forward(self, tokens: torch.Tensor) -> torch.Tensor:
        """Forward pass through token embedding module.
        :param tokens: Tokens to embed
        :type tokens: torch.Tensor
        """
        # return self.embedding(tokens.long()) * math.sqrt(self.emb_size)
        return self.embedding(tokens)


def _none_or_dtype(input: Any) -> Optional[torch.dtype]:
    """Returns None or dtype of input. Adapted from torch.nn.functional.
    :param input: Input to return dtype of
    :type input: Any
    """
    if input is None:
        return None
    elif isinstance(input, torch.Tensor):
        return input.dtype
    raise RuntimeError("input to _none_or_dtype() must be None or torch.Tensor")
